#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


# ==========================================================================
# set up the load path
#
import sys
import os.path

path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.append(path)
# ==========================================================================


import gtk.glade
import gtk.gdk
import cairo
import pygtk
import os.path
import re
import itertools
import cgi # for cgi.escape
from math import pi
from StringIO import StringIO
import operator as op

import sevendown.loader as sl
import sevendown.solver as ss
import sevendown.logger
import sevendown.graph
import sevendown.pretty
import sevendown.widgets as sw
import sevendown.undo as undo

# State variables. The GUI has always a certain state,
# and this should be set trough set_state.
STATE_NONE, STATE_FILE_LOADED, STATE_EDITING, STATE_RUNNING, STATE_PAUSED, \
        STATE_STEPPING, STATE_DONE = range(7)
state_texts = [ "Program started",
                "File loaded",
                "Editing source",
                "Running solver",
                "Solver paused",
                "Solver stepping",
                "Solver done"]

class StopSolver(Exception):
    """
    Exception to stop the Solver. This is used in the Solvers’ callback
    functions, because it would continue to run if we just returned here.
    """
    pass

class Gui(sevendown.logger.NullLogger):
    """ main class for gui application. """

    def __init__(self):
        self.tree = gtk.glade.XML(os.path.join(
            os.path.dirname(__file__), '..', '..', 'data', '7down.glade'))
        self.mainWindow = self.tree.get_widget("7down")
        self.openFileDialog = self.tree.get_widget("filechooserdialog")
        self.saveFileDialog = self.tree.get_widget("filechooserdialog_save")
        self.logTextView = self.tree.get_widget("textviewLog")
        self.rulesTreeView = self.tree.get_widget("rulesTreeView")
        self.sourceTextView = self.tree.get_widget("sourceTextView")
        
        self.fieldAspectFrame = self.tree.get_widget('fieldAspectFrame')
        self.fieldWidget = sw.FieldWidget()
        self.fieldAspectFrame.add(self.fieldWidget)
        self.fieldWidget.show()
        self.fieldWidget.connect('digit-clicked', self.on_fieldwidget_digit_clicked)
        
        self.tree.signal_autoconnect(self)
        self.setup_rules_treeview()

        self.problem = None
        self.solver = None
        self.fieldWidget.setHighlights({})
        self.filename = "untitled.in"

        self.state = None
        self.set_state(STATE_NONE)

        self.set_field_ratio()

        # We want the solver to signal us relevant changes
        sevendown.logger.logger = self

        # Nice default path for the file dialogue
        self.openFileDialog.set_current_folder(os.path.join(
            os.path.dirname(os.path.abspath(__file__)), '..', '..', 'tests', 'examples'))

        # If we have a parameter, try to open it as a file
        if len(sys.argv)>1:
            self.load_file(os.path.abspath(sys.argv[1]))
            self.parse_source()

    def set_state(self, state):
        """
        This function sets the object’s state, and updates various
        parts of the GUI reflecting that state.
        """
        self.state = state
        self.tree.get_widget('buttonStep').props.sensitive = \
        self.tree.get_widget('buttonStart').props.sensitive = \
            state in (STATE_FILE_LOADED, STATE_PAUSED, STATE_DONE)
        self.tree.get_widget('buttonOpen').props.sensitive = \
            state in (STATE_FILE_LOADED, STATE_EDITING, STATE_NONE, STATE_DONE)
        self.tree.get_widget('buttonStop').props.sensitive = \
            state in (STATE_RUNNING, STATE_PAUSED, STATE_DONE)
        self.tree.get_widget('buttonPause').props.sensitive = \
            state in (STATE_RUNNING,)
        self.tree.get_widget('textviewLog').props.sensitive = \
            state not in (STATE_RUNNING,)
        self.tree.get_widget('sourceEditable').props.sensitive = \
            state in (STATE_NONE, STATE_FILE_LOADED, STATE_EDITING, STATE_DONE)
        self.tree.get_widget('sourceTextView').props.editable = \
        self.tree.get_widget('sourceEditable').props.active = \
            state in (STATE_NONE, STATE_EDITING,)
            
        # set window title
        self.mainWindow.set_title("7down − %s − %s" % (state_texts[self.state], self.filename))
        self.stateWasUpdated()
    
    def niceList(self,l):
        return ", ".join(map(str,l))

    def stateWasUpdated(self):
        # List Number Values and assumptions
        if self.solver:
            ass = ""
            if not self.solver.splits_to_try:
                ass += "\tNone yet"
            else:
                for c, others in self.solver.splits_to_try:
                    if others:
                        others_text = "will try " + self.niceList(reduce(op.or_,others,set())) + " later"
                    else:
                        others_text = "nothing else left to try"
                    ass += "\tAssume %s contains %s, and %s.\n" % (
                            str(c),
                            self.niceList(c.getFeasibleValues()),
                            others_text
                            )
            
            buf = self.tree.get_widget("numberView").get_buffer()
            buf.set_text("Assumptions:\n%s\nNumbers:\n" % ass)
            
            table = buf.get_tag_table()
            for cell in self.problem.cells:
                if isinstance(cell, sevendown.graph.Number) and not isinstance(cell, sevendown.graph.Fixed):
                    color = self.fieldWidget.highlights.get(cell, (1,1,1))
                    tag = table.lookup(str(color))
                    if not tag:
                        tag = buf.create_tag(str(color),
                                background_gdk = self.color_tuple_to_gdk(*color))
                    iter = buf.get_end_iter()
                    buf.insert_with_tags(iter,
                            sevendown.pretty.CellPrinter(cell).run(),
                            tag)
        
        # redraw widget
        self.fieldWidget.queue_draw()
    
    def on_buttonExit_clicked(self, widget, data=None):
        self.mainWindow.destroy()
    
    def on_buttonOpen_clicked(self, widget, data=None):
        self.openFileDialog.props.visible = True
    
    def on_buttonSave_clicked(self, widget, data=None):
        if self.filename == "untitled.in" or widget.name == 'buttonSaveAs':
            self.saveFileDialog.set_filename(self.filename)
            response = self.saveFileDialog.run()
            if response == gtk.RESPONSE_OK:
                self.filename = self.saveFileDialog.get_filename()
                self.saveFileDialog.hide()
            else:
                # user aborted
                self.saveFileDialog.hide()
                return

        f = file(self.filename, 'w')
        buf = self.sourceTextView.get_buffer()
        text = buf.get_text(*buf.get_bounds())
        f.write(text)

    def on_fieldwidget_digit_clicked(self, widget, shift, cell, digit):
        # Simple editing in the pre-load-phase
        if self.state in (STATE_FILE_LOADED,):
            if shift:
                cell.force(set([digit]))
            else:
                feasible_values = cell.getFeasibleValues()
                if digit in feasible_values:
                    cell.force(feasible_values ^ set([digit]))
                else:
                    cell.force(feasible_values | set([digit]))

        if self.state in (STATE_RUNNING, STATE_PAUSED):
            feasible_values = cell.getFeasibleValues()
            if digit in feasible_values and len(feasible_values)>1:
                if shift:
                    # put other digits on stack:
                    undo.startRevision()
                    cell.force(set([digit]))
                    self.solver.splits_to_try.append((cell, [feasible_values ^ set([digit])]))
                else:
                    # put digit on stack to try
                    undo.startRevision()
                    cell.force(feasible_values ^ set([digit]))
                    self.solver.splits_to_try.append((cell, [set([digit])]))
                self.solver.just_splitted_cells.append(cell)
            
        self.stateWasUpdated()

    def on_notebook_switch_page(self, widget, page, page_num):
        # Save buttons
        source_tab_num = widget.page_num(self.tree.get_widget('sourceVBox'))
        self.tree.get_widget('buttonSave').props.visible = \
        self.tree.get_widget('buttonSaveAs').props.visible = \
            page_num == source_tab_num


    def on_7down_destroy(self, widget, data=None):
        sys.exit()

    def on_7down_delete(self, widget, data=None):
        return False

    def on_filechooserdialog_response(self, filechooser, response, data=None):
        """
        If the user selected a file, try to open it. Unless that fails, 
        hide the openFileDialogue.
        """
        if response == gtk.RESPONSE_OK:
            if self.load_file(filechooser.get_filename()):
                self.openFileDialog.hide()
                # Be smart here: If the edit notepad is visible
                # then just load the file, otherwise parse it.
                # If parsing fails, stay in edit mode and open the notepad
                notebook = self.tree.get_widget('notebook')
                source_tab_num = notebook.page_num(self.tree.get_widget('sourceVBox'))
                if notebook.get_current_page() != source_tab_num:
                    # The use does not have the source tab open
                    if not self.parse_source():
                        # But parsing failed, select the source tab.
                        notebook.set_current_page(source_tab_num)
        else:
            self.openFileDialog.hide()
    
    def on_buttonStart_clicked(self, widget, data=None):
        if self.state == STATE_PAUSED:
            # we have just paused, so make the current gtk main loop quit
            gtk.main_quit()
            self.set_state(STATE_RUNNING)
        else:
            if self.state == STATE_DONE:
                self.parse_source()
            self.run()

    def on_buttonStop_clicked(self, widget, data=None):
        """
        This does not yet really stop, but signals the program to stop
        at the next possible point.
        """
        if self.state == STATE_PAUSED:
            # we have just paused, so make the current gtk main loop quit
            gtk.main_quit()
        elif self.state == STATE_DONE:
            # now its some kind of reset button
            self.parse_source()
        self.set_state(STATE_DONE)

    def on_buttonStep_clicked(self, widget, data=None):
        """
        This does not yet really stop, but signals the program to stop
        at the next possible point.
        """
        if self.state == STATE_PAUSED:
            # we have just paused, so make the current gtk main loop quit
            gtk.main_quit()
            self.set_state(STATE_STEPPING)
        else:
            if self.state == STATE_DONE:
                self.parse_source()
            self.run(new_state=STATE_STEPPING)

    def on_buttonPause_clicked(self, widget, data=None):
        """
        This does not yet really stop, but signals the program to stop
        at the next possible point.
        """
        assert self.state == STATE_RUNNING
        self.set_state(STATE_PAUSED)
        gtk.main()

    def on_rulesTreeView_cursor_changed(self, widget, data=None):
        self.rule_selected()

    def on_sourceEditable_toggled(self, widget, data=None):
        if widget.props.active:
            # User wants to edit the source
            # Maybe we should warn him that this resets the solving procedure,
            # if we are in the running or paused state.
            self.set_state(STATE_EDITING)

        else:
            # User wants to stop editing the source
            self.parse_source()

    def load_file(self, filename):
        """
        Loads a file into the source edit.

        Returns False if the loading has failed.
        """
        try:
            self.sourceTextView.get_buffer().set_text(file(filename).read())
            self.filename = filename
            self.set_state(STATE_EDITING)
            return True
        except Exception, e:
            # Someone please check the correct exception to handle here.
            print e
            return False

    def parse_source(self):
        """
        Parses the source in the source edit into self.problem,
        and initializes self.parser.
        """
        try:
            buf = self.sourceTextView.get_buffer()
            source = buf.get_text(*buf.get_bounds())
            loader = sl.Loader()
            self.problem = loader.load(StringIO(source))
            self.solver = ss.StackSolver()
            self.solver.setProblem(self.problem)
            self.fieldWidget.set_field_and_solver(self.problem, self.solver)

            self.set_field_ratio()

            self.load_rules()

            self.set_state(STATE_FILE_LOADED)
            self.clearLog()
            self.appendLog("Sucessfully loaded problem %s.\n" % self.filename)

            return True
            
        except (sl.ParserException, ValueError), e:
            # ValueError happens when a split does not
            # return the right number of arguments.
            dialog = gtk.MessageDialog(self.mainWindow, gtk.DIALOG_MODAL,
                    gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, str(e))
            dialog.run()
            dialog.destroy()
            
            # Reset all visible state
            self.solver = None
            self.problem = None
            self.fieldWidget.setHighlights({})
            self.set_field_ratio()
            self.rulesTreeView.set_model(None)
            self.rule_selected()
            self.set_state(STATE_EDITING)

            return False

    def setup_rules_treeview(self):
        # Anonymous function to format cell
        def set_text(column, cell, model, iter):
            cell.props.text = str(model.get_value(iter, 0))
        self.rulesTreeView.insert_column_with_data_func(
                -1,
                "Rule",
                gtk.CellRendererText(),
                set_text
                )
        self.rule_selected()

    def load_rules(self):
        model = gtk.ListStore(object)
        for rule in self.problem.rules:
            model.append([rule])
        self.rulesTreeView.props.model = model

    def rule_selected(self):
        details = self.tree.get_widget("ruleDetailsLabel")
        path, _ = self.rulesTreeView.get_cursor()
        model   = self.rulesTreeView.props.model
        if path and model:
            # We have selected something
            iter = model.get_iter(path)
            rule = model.get_value(iter,0)

            rule_details = rule.detailedDescription()
            text = "%s:\n%s" % (cgi.escape(str(rule)), cgi.escape(rule_details))

            # Highlight relevant cells here
            highlights = {}
            cells = filter(lambda c: not isinstance(c,sevendown.graph.Fixed), rule.relatedCells())
            for cell, color in zip(cells, self.highlight_colors()):
                highlights[cell] = color
                repl = '<span background="%s">%s</span>' % (
                        self.color_tuple_to_pango(*color), cgi.escape(str(cell))
                        )
                text = text.replace(cgi.escape(str(cell)), repl)
            self.fieldWidget.setHighlights(highlights)
                    
            details.set_markup(text)
            self.stateWasUpdated()

        else:
            # Nothig is selected or selectable
            details.set_markup("<i>(keine Regel ausgewählt)</i>")

    def run(self, new_state=STATE_RUNNING):
        """
        This actually runs the solver.
        It calls gtk.main_iteration from time to time,
        so that the program stays responsive.
        """
        # Nested try because python2.4 does not support
        # the try: .. except: .. finally: scheme
        try:
            try:
                self.solver.start()
                self.set_state(new_state)

                while not self.solver.step():
                    self.stateChangeOnCallback()

            finally:
                # We need to to this, otherwise the undo manger
                # will break on the next run.
                result = self.solver.stop()
        except StopSolver:
            # This exception is just to abort the solver, don’t pass 
            # it to the outside world
            result = None
            pass

        if result is not None:
            # Solver has stopped by itself,
            if result:
                text = "Problem solved."
            else:
                text = "No solution found."
            dialog = gtk.MessageDialog(self.mainWindow, gtk.DIALOG_MODAL,
                    gtk.MESSAGE_INFO, gtk.BUTTONS_OK, text)
            dialog.run()
            dialog.destroy()

        self.stateWasUpdated()
        self.set_state(STATE_DONE)

    def clearLog(self):
        """
        Erases the log view window
        """
        buf = self.logTextView.get_buffer()
        buf.set_text("")

    def appendLog(self, text):
        """
        This appens some text to the log viewer and scrolls it down.
        """
        buf = self.logTextView.get_buffer()
        iter = buf.get_end_iter()
        buf.insert(iter, text)
        self.logTextView.scroll_mark_onscreen(buf.get_insert())

    def stateChangeOnCallback(self, stop_stepping=True):
        '''
        This must be called whenever possible while running the solver, 
        to react on state changes.
        '''
        self.stateWasUpdated()
        while gtk.events_pending():
            gtk.main_iteration(block = False)

        if self.state == STATE_DONE:
            raise StopSolver()
        if stop_stepping and self.state == STATE_STEPPING:
            # we are stepping,so wait for user input
            self.set_state(STATE_PAUSED)
            gtk.main()

    #
    # Callback functions from the solver
    #
    def applying(self, rule):
        pass

    def ruleResult(self, rule, res, field):
        if res:
            highlights = {}
            msg = "Rule %s changed cells:\n" % rule
            for cell, color in zip(res, self.highlight_colors()):
                highlights[cell] = color
            self.fieldWidget.setHighlights(highlights)
                    
            for cell in res:
                if isinstance(cell, sevendown.graph.Number):
                    msg += "\t" + sevendown.pretty.CellPrinter(cell).run()
                else:
                    msg += "\t%s\n" % cell
            self.appendLog(msg)

        self.stateChangeOnCallback(stop_stepping = bool(res))

    def contradictionOccured(self, rule, cell):
        tpl = "Contradiction occured for cell %s after applying rule %s\n"
        msg = tpl % (cell, rule)
        self.appendLog(msg)
    
        self.stateChangeOnCallback()

    def applyingSplit(self, cell, split):
        self.appendLog("Splitting %s to %s\n" % (cell, split))
    
    def consideringSplits(self, cell, splits):
        self.appendLog("split returned %s\n" % (splits))
    
        self.stateChangeOnCallback()

    # Drawing functions
    
    def set_field_ratio(self):
        self.fieldAspectFrame.props.ratio = \
            float(self.width()) / float(self.height())
    
    def width(self):
        """ Number of cells horizontally"""
        if self.problem:
            return self.problem.width
        else:
            return 8
    
    def height(self):
        """ Number of cells vertically"""
        if self.problem:
            return self.problem.height
        else:
            return 8
    
    def color_tuple_to_gdk(self, r, g, b):
        return gtk.gdk.Color(int(r * 65535), int(g * 65535), int(b * 65535))
    
    def color_tuple_to_pango(self, r, g, b):
        return "#%02X%02X%02X" % (r*0xFF, g*0xFF, b*0xFF)

    def highlight_colors(self):
        return itertools.cycle(
                    (   (1,   0.5, 0.5),
                        (0.5, 1,   0.5),
                        (0.5, 0.5, 1   )))
    
    def main(self):
        gtk.main()

if __name__ == "__main__":
    Gui().main()

# vim:sw=4:ts=4:sts=4:expandtab 
