#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""This module defines the Solver and related classes."""

import operator
import random
import heapq

import sevendown.config as sc
import sevendown.graph as sg
import sevendown.rules as sr
import sevendown.pretty as sp
import sevendown.statistics as ss
import sevendown.undo as undo
import sevendown.logger as l


def max(l, key=(lambda x:x)):
    assert len(l)>0, "Max on empty list"
    if type(l) == dict:
        l = l.keys()
    m = l[0]
    for e in l:
        if key(e) > key(m):
            m = e
    return m

class SolverNotApplicable(Exception):
    pass

class RulesHeapItem(object):
    """
    Wrapper for (rules, set of related cells) pairs.  Ordered by the rule.
    
    """
    
    def __init__(self, rule, cells):
        self.rule = rule
        self.cells = cells
    
    def get_rule(self):
        return self.rule
    
    def get_cells(self):
        return self.cells
    
    def __cmp__(self, other):
        assert isinstance(self, RulesHeapItem)
        return cmp(self.rule, other.rule)
    

class NextSplittableSelector(object):
    """
    Container class for 'select next splittable cell' logic.
    
    TODO: This might eventually be more pythonesque if it is converted into
    TODO: module level methods.
    
    """
    
    @classmethod
    def select(self, cells):
        """
        Returns the index of a splittable cell if any and None otherwise.
        
        The method which is used depends on the configuration setting
        sevendown.config.SOLVER_NEXT_SPLITTABLE.
        
        """
        methods = {sc.SOLVER_NEXT_SPLITTABLE_FIRST:self.selectFirst,
            sc.SOLVER_NEXT_SPLITTABLE_SMALLEST:self.selectSmallest,
            sc.SOLVER_NEXT_SPLITTABLE_LARGEST:self.selectLargest,
            sc.SOLVER_NEXT_SPLITTABLE_RANDOM:self.selectRandom,
            sc.SOLVER_NEXT_SPLITTABLE_MOST_RULES:self.selectWithMostRules}
        return methods[sc.SOLVER_NEXT_SPLITTABLE](cells)
    
    @classmethod
    def selectFirst(self, cells):
        """Returns the index of the first cell which is splittable."""
        for (i, cell) in enumerate(cells):
            if cell.isSplittable():
                return i
        return None
    
    @classmethod
    def selectLargest(self, cells):
        """Returns the index with the largest cardinality."""
        return self._selectSmallestLargestHelper(-1, 
            operator.gt, cells)
        
    
    @classmethod
    def selectSmallest(self, cells):
        """Returns the index with the smallest cardinality."""
        return self._selectSmallestLargestHelper(sg.INFINITY, 
            operator.lt, cells)
    
    @classmethod
    def _selectSmallestLargestHelper(self, initial, comp, cells):
        """Helper for selectSmallest, selectLargest methods."""
        if not cells: # if empty
            return None
        
        index, card = None, initial
        
        for i, cell in enumerate(cells):
            if cell.isSplittable():
                if comp(cell.cardinality(), card):
                    index = i
                    card = cell.cardinality()
        
        return index
    
    @classmethod
    def selectRandom(self, cells):
        """Returns a random splittable cell from cells."""
        result = random.randrange(len(cells))
        while not cells[result].isSplittable():
            result = random.randrange(len(cells))
        
        return result
    
    @classmethod
    def selectWithMostRules(self, cells):
        """Returns the cell with the most related rules."""
        if not cells: # if empty
            return None
        
        index, count = None, -1
        
        for i, cell in enumerate(cells):
            if cell.isSplittable():
                if len(cell.getRelatedRules()) > count:
                    index = i
                    count = len(cell.getRelatedRules())
        
        return index
    

class Solver(object):
    """Solver base class."""
    
    def __init__(self):
        self.problem = None
    
    def isDone(self):
        """Returns True iff all cells of the problem return True on isDone()
        and all rules are satisfied.
        
        Returns False otherwise.
        
        """
        for cell in self.problem.cells:
            if not cell.isDone():
                l.logger.notDoneBecauseOf(cell)
                return False
        
        return True
    
    def setProblem(self, problem):
        """Sets the problem.
        """
        self.problem = problem
    
    def solve(self, problem):
        self.setProblem(problem)
        if self.runSolver():
            return self.problem
        else:
            return None
    
    def selectSplittableIndex(self):
        """
        Returns the index of the next splittable cell.
        
        This method the NextSplittableSelector class to perform this
        task.  This object will use the configuration setting 
        SOLVER_NEXT_SPLITTABLE to select between different behaviours.
        
        """
        return NextSplittableSelector.select(self.problem.cells)
    

class OrderedSolver(Solver):
    """Tries to run exactly those roles that need be re run.
    
    Requires Digit.getRelatedRules()
    """
    
    def __init__(self, **args):
        Solver.__init__(self, **args)
        self.statistics = ss.Statistics()
        
        self.rules_to_try = set()
    
    def setProblem(self, problem):
        """Sets the problem and rules to try."""
        self.problem = problem
        
        self.rules_to_try = set(self.problem.rules)
    
    def runSolver(self):
        """Runs the solver on the problem in self.problem.
        
        Returns True iff the solver could solve the problem.  In this case,
        self.problem contains the solved field.  Only returns the first
        solution found by the Solver.
        
        Override parent class' method to implement custom behaviour.
        
        """
        # Start the first revision, never to be rolled back:
        undo.setUp()
        
        # print rules ONCE
        l.logger.startingSolver(self.problem)
        
        # Now, we can try to solve the problem recursively.
        try:
            result = self.runSolverRecursion()
        finally:
            undo.tearDown()
        return result
    
    def runSolverRecursion(self):
        """Called to solve the problem recursively."""
        self.statistics.registerRecursion()
        l.logger.solvingStep(self.problem)
        
        # Rule applying phase
        while self.rules_to_try:
            rule = self.rules_to_try.pop()
            
            l.logger.applying(rule)
            
            res = rule.apply()
            self.statistics.registerOrdinaryRuleApplication(rule)
            
            l.logger.ruleResult(rule, res, self.problem)
            
            for cell in res:
                if cell.isEmpty():
                    l.logger.contradictionOccured(rule, cell)
                    self.statistics.registerReturnFromRecursion()
                    return False
                else:
                    for rule2 in cell.getRelatedRules():
                        if rule2 != rule:
                            self.rules_to_try.add(rule2)
        
        # At this point, if some rule failed, we detected that and
        # have returned. Therefore we can assume that all rules are
        # satisfiable.
        
        # Are we done directly after splitting and recursing?
        if self.isDone():
            self.statistics.registerReturnFromRecursion()
            return True
        
        l.logger.solvingStep(self.problem)
        
        # We are not done, but tried all rules. Time to brute force:
        
        # Apply all rules until no rule can be applied any more.
        i = self.selectSplittableIndex()
        if i == None:
            # No solution, but also no further splitting possible.
            self.statistics.registerReturnFromRecursion()
            return False
        
        # Split next cell if necessary.
        splits = self.problem.cells[i].split()
        l.logger.consideringSplits(self.problem.cells[i], splits)
        
        for split in splits:
            assert len(split) > 0, "split() returned an emty set"
            
            undo.startRevision()
            l.logger.applyingSplit(self.problem.cells[i], split)
            
            self.problem.cells[i].force(split)
            
            # Recurse, trying cell-related rules
            
            undo.registerAttribute(self, "rules_to_try")
            if not self.problem.cells[i].getRelatedRules():
                print "Cell %s has no related rules!" % str(self.problem.cells[i])
            for rule in self.problem.cells[i].getRelatedRules():
                l.logger.willTryRule(rule)
                self.rules_to_try.add(rule)
            
            ret = self.runSolverRecursion()
            if ret:
                self.statistics.registerReturnFromRecursion()
                return True
            
            undo.revert()
        # ==================================================================
        
        # No split was successful: fail
        self.statistics.registerReturnFromRecursion()
        return False
    

class StackSolver(Solver):
    """Tries to run exactly those rules that need be re run.
    
    Uses explicit stepping.
    """
    
    def __init__(self, **args):
        Solver.__init__(self, **args)
        self.statistics = ss.Statistics()
        
        self.just_splitted_cells = []
        
        # A list of (cell,[split]) tuples
        self.splits_to_try = []
    
    def setProblem(self, problem):
        """Sets the problem and rules to try."""
        self.problem = problem
        self.just_splitted_cells = []
        
    
    def runSolver(self):
        """Runs the solver on the problem in self.problem.
        
        Returns True iff the solver could solve the problem.  In this case,
        self.problem contains the solved field.  Only returns the first
        solution found by the Solver.
        
        Override parent class' method to implement custom behaviour.
        
        """
        
        self.start()
        while not self.step():
            pass
        return self.stop()
        # try:
        #     while not self.step():
        #         pass
        # finally:
        #     return self.stop()
    
    def start(self):
        """
        Starts the stepping process
        """
        # Start the first revision, never to be rolled back:
        undo.setUp()
        
        self.splits_to_try = []
        
        # print rules ONCE
        l.logger.startingSolver(self.problem)
    
    def step(self):
        """
        Does one step. Returns False if it wants to be called again
        
        """
        if self.splits_to_try:
            if not self.splitStep():
                # done, unsolvable
                return True
        
        done = self.ruleStep()
        
        if done or not self.splits_to_try:
            return True
        else:
            return False
    
    def stop(self):
        """
        Finishes the stepping run. Returns True if solution was found
        """
        undo.tearDown()
        
        # If there is nothing left to try, then we have not found a solution
        return self.isDone()
    
    def ruleStep(self):
        l.logger.solvingStep(self.problem)
        
        rules_to_try = []
        related_cells = {} # {rule: set(<related cells>)}
        
        if self.just_splitted_cells:
            while self.just_splitted_cells:
                cell = self.just_splitted_cells.pop()
                for rule in cell.getRelatedRules():
                    l.logger.willTryRule(rule)
                    
                    if not rule in rules_to_try: heapq.heappush(rules_to_try, rule)
                    related_cells.setdefault(rule, set()).add(cell)
        else:
            all_cells = set(self.problem.cells)
            for rule in self.problem.rules:
                if not rule in rules_to_try: heapq.heappush(rules_to_try, rule)
                related_cells.setdefault(rule, all_cells)
        
        # print "rules to try: %s" % rules_to_try
        
        # Rule applying phase
        while rules_to_try:
            # get rule with highest priority from set (TODO: Improve O(n))
            # print rule
            rule = heapq.heappop(rules_to_try)
            causing_cells = related_cells.pop(rule)
            
            l.logger.applying(rule)
            
            res = rule.apply(causing_cells)
            self.statistics.registerOrdinaryRuleApplication(rule)
            
            l.logger.ruleResult(rule, res, self.problem)
            
            for cell in res:
                if cell.isEmpty():
                    l.logger.contradictionOccured(rule, cell)
                    undo.revert()
                    return False
                else:
                    for rule2 in cell.getRelatedRules():
                        if rule2 != rule:
                            l.logger.willTryRule(rule2)
                            if not rule2 in rules_to_try: heapq.heappush(rules_to_try, rule2)
                            related_cells.setdefault(rule2, set()).add(cell)
        
        # At this point, if some rule failed, we detected that and
        # have returned. Therefore we can assume that all rules are
        # satisfiable.
        
        # Are we done directly after splitting and recursing?
        if self.isDone():
            return True

        # We are not done, but tried all rules. Add new splits to do

        i = self.selectSplittableIndex()
        if i == None:
            # No solution, but also no further splitting possible.
            undo.revert()
            return False
        
        # Split next cell if necessary.
        splits = self.problem.cells[i].split()[::-1]
        assert len(splits) > 0
        l.logger.consideringSplits(self.problem.cells[i], splits)
        self.splits_to_try.append((self.problem.cells[i], splits))
        return False

    def splitStep(self):
        """
        Execute 'split step'.  Returns False if splitting is impossible and
        problem is unsolvable.  Otherwise, True is returned and the problem
        can maybe still be solved.
        
        """
        # skip split step if there is nothing to split
        if not self.splits_to_try:
            return False
        
        l.logger.solvingStep(self.problem)
        
        if self.splits_to_try[-1][1]:
            cell = self.splits_to_try[-1][0]
            split = self.splits_to_try[-1][1].pop()
            
            assert len(split) > 0, "split() returned an empty set"
            
            l.logger.applyingSplit(cell, split)
            
            undo.startRevision()
            self.statistics.registerRecursion()
            
            cell.force(split)
            self.just_splitted_cells.append(cell)
            
            return True
        else:
            self.splits_to_try.pop()
            undo.revert()
            
            return self.splitStep()
    

# vim:sw=4:ts=4:sts=4:expandtab
