#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Utility code.

"""

import math

def combine(*seqin):
    """
    Returns a list of all combinations of argument sequences.
    for example: combine((1,2),(3,4)) returns
    [[1, 3], [1, 4], [2, 3], [2, 4]]
    
    Taken from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/302478
    
    """
    def rloop(seqin,listout,comb):
        """Recursive looping function."""
        if seqin:                       # any more sequences to process?
            for item in seqin[0]:
                newcomb=comb+[item]     # add next item to current comb
                # call rloop w/ rem seqs, newcomb
                rloop(seqin[1:],listout,newcomb)
        else:                           # processing last sequence
            listout.append(comb)        # comb finished, add to list
    listout=[]                      # listout initialization
    rloop(seqin,listout,[])         # start recursive process
    return listout


def xcombine(seqin):
    """
    Returns a generator which returns combinations of argument sequences
    for example xcombine((1,2),(3,4)) returns a generator; calling the next()
    method on the generator will return [1,3], [1,4], [2,3], [2,4] and
    StopIteration exception.  This will not create the whole list of
    combinations in memory at once.
    
    Taken from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/302478
    
    """
    def rloop(seqin,comb):
        """recursive looping function"""
        if seqin:                   # any more sequences to process?
            for item in seqin[0]:
                newcomb=comb+[item]     # add next item to current combination
                # call rloop w/ remaining seqs, newcomb
                for item in rloop(seqin[1:],newcomb):
                    yield item          # seqs and newcomb
        else:                           # processing last sequence
            yield comb                  # comb finished, add to list
    return rloop(seqin,[])


def to_int(seq):
    return map(assembleDigits, seq)


def assembleDigits(seq):
    v = 0
    for s in seq:
        v *= 10
        v += s
    return v


def digits(n, l=None):
    """
    Returns a list of digits in the number, padded to lenght l with zeros.
    
    If l is given and number does not fit, returns None.

    For zero, returns zeros. For negative parameters, returns the digits of the
    absolute value.
    
    """
    if l is None:
        l = len(str(n))

    d = [0]*l
    
    if n == 0:
        return d

    if n < 0:
        n = -n
    
    m = n
    for i in range(1,l+1):
        d[-i] = m % 10
        m = m // 10
    if m != 0:
        return None
    else:
        return d


def pot_numbers(pot, begin, end):
    """
    Returns a generator for all pot potences in [begin, end].
    """
    
    begin = int(math.floor(math.pow(begin, 1.0/pot)))
    end = int(math.ceil(math.pow(end, 1.0/pot)))
    
    return (x**pot for x in xrange(begin, end))


def xmerge(*ln):
     """
     Iterator version of merge.
     
     Assuming l1, l2, l3...ln sorted sequences, return an iterator that
     yield all the items of l1, l2, l3...ln in ascending order.
     Input values doesn't need to be lists: any iterable sequence can be used.
     
     Source: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/141934
     
     """
     pqueue = []
     for i in map(iter, ln):
         try:
             pqueue.append((i.next(), i.next))
         except StopIteration:
             pass
     pqueue.sort()
     pqueue.reverse()
     X = max(0, len(pqueue) - 1)
     while X:
         d,f = pqueue.pop()
         yield d
         try:
             # Insort in reverse order to avoid pop(0)
             lo, hi, d = 0, X, f()
             while lo < hi:
                 mid = (lo+hi)//2
                 if d > pqueue[mid][0]: hi = mid
                 else: lo = mid+1
             pqueue.insert(lo, (d,f))
         except StopIteration:
             X-=1
     if pqueue:
         d,f = pqueue[0]
         yield d
         try:
             while 1: yield f()
         except StopIteration:pass


def merge(*ln):
     """
     Merge several sorted sequences into a sorted list.
     
     Assuming l1, l2, l3...ln sorted sequences, return a list that contain
     all the items of l1, l2, l3...ln in ascending order.
     Input values doesn't need to be lists: any iterable sequence can be used.
     
     Source: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/141934
     
     """
     return list(xmerge(*ln))


def identity(arg):
    """The identity function; sometimes it comes in handy."""
    return arg


def iterable2string(iterable):
    """
    Converts an iterable (lists, sets) into a string containing the comma 
    separated values.
    
    """
    return ",".join(map(str, iterable))


FACTOR_LIST = None

def factors_of(num):
    """
    Returns a set with all factors of num (modulo 10).
    
    """
    assert (num >= 0) and (num <= 9)
    
    global FACTOR_LIST
    if FACTOR_LIST is None:
        FACTOR_LIST = [set() for i in range(10)]
        
        for i in range(0, 10):
            for j in range(1, 10):
                x = i * j
                m = x % 10
                FACTOR_LIST[m].add(i)
                FACTOR_LIST[m].add(j)
    
    return FACTOR_LIST[num]


def gcd(a,b):
    """Return greatest common divisor using Euclid's Algorithm."""
    while b:      
        a, b = b, a % b
    return a


def gcdList(list):
    """Return greatest common divisor across the list"""
    assert len(list)>0, "What's the gcd of nothing?"
    return reduce(gcd, list)


# vim:sw=4:ts=4:sts=4:expandtab
