#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import unittest
import test_utils
import StringIO

import sevendown.rules
import sevendown.rules as sr
import sevendown.loader as sl
import sevendown.graph
import sevendown.graph as sg
import sevendown.undo as undo

from todo import TODO

MINIMAL_EXAMPLE = """2 2
2
A 1 1
B 1 2
0
2
A
B
1
PZ(A_s)
"""

class LoaderTestCase(test_utils.TestCase):
    """Tests for the class Rule."""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.loader = sl.Loader()
    
    def testLoadDimensions(self):
        in_io = StringIO.StringIO("1 2")
        expected = (1, 2)
        
        result = self.loader.loadDimensions(in_io)
        
        self.failUnlessEqual(result, expected)
    
    def testLoadMarkerCoordinates(self):
        
        self.loader.height = 2
        self.loader.width = 3
        
        in_io = StringIO.StringIO("2\nA 1 1\nB 2 3")
        expected = {'A': ('A', 0, 0), 'B':('B', 2, 1)}
        
        result = self.loader.loadMarkerCoordinates(in_io)
        
        self.failUnlessEqual(result, expected)
    
    def testLoadMarkerCoordinatesWithFail(self):
        """Test that invalid markers raise exception."""
        
        self.loader.height = 2
        self.loader.width = 2
        in_io = StringIO.StringIO("2\nA 1 1\nB 2 3")
        
        def call():
            self.loader.loadMarkerCoordinates(in_io)
        
        self.assertRaises(sl.MarkerOutOfFieldException, call)
    
    def testLoadMarkerNames(self):
        in_io = StringIO.StringIO("2\nA\nB")
        expected = ('A', 'B')
        
        result = tuple(self.loader.loadMarkerNames(in_io))
        
        self.failUnlessEqual(result, expected)
    
    def testLoadRuleLines(self):
        in_io = StringIO.StringIO("2\nPZ(A_w)\n=(A_w, B_s)")
        expected = ('PZ(A_w)', '=(A_w, B_s)')
        
        result = tuple(self.loader.loadRuleLines(in_io))
        
        self.failUnlessEqual(result, expected)
    
    def testLoadSimpleExample(self):
        """Tests the loader with a simple, minimal example.
        
        """
        # write minimal example to buffer
        in_io = StringIO.StringIO(MINIMAL_EXAMPLE)
        
        # loader from buffer
        problem = self.loader.load(in_io)
        
        self.assertEqual(len(problem.rules), 6)
        self.assertEqual(len(problem.numbers), 2)
    

class PredicateParserTestCase(test_utils.TestCase):
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.parser = sl.PredicateParser()
    
    def testTokenize(self):
        """Test the method tokenize() by passing in some values.
        
        """
        values = (
            ("=", "(", "qs", "(", "A_w", ")", ",", " 14", ")"),
            ("PZ", "(", "F_w", ")"),
            ("=", "(", "G_w", ",", "minus", "(", "mal", "(", "6", ",", "H_w",
                ")", ",", "B_s", ")", ")"),
            ("Q", "(", "H_w", ")"),
            ("=", "(", "I_w", ",", "r", "(", "F_w", ")", ")"),
            ("V", "(", "J_w", ",", "F_w", ")"),
            ("Q", "(", "M_w", ")"),
            ("V", "(", "N_w", ",", "3", ")"),
            ("=", "(", "A_s", ",", "mal", "(", "B_s", ",", " H_w", ")", ")"),
            ("PAL", "(", "B_s", ")"),
            ("V", "(", "C_s", ",", "D_s", ")"),
            ("PZ", "(", "r", "(", "D_s", ")", ")"),
            ("=", "(", "r", "(", "E_s", ")", ",", "minus", "(", "A_w", ",",
                "10", ")", ")"),
            ("P", "(", "K_s", ")"),
            ("Q", "(", "L_s", ")")
        )
        
        for value in values:
            the_input = "".join(value)
            result = self.parser.tokenize(the_input)
            
            # commas and whitespace are not returned
            expected = tuple(token.strip() for token in value if token != ',')
            
            self.failUnlessEqual(result, expected)
    
    def testParse(self):
        """Test the parser."""
        tokens = ("=", "(", "A_s", "mal", "(", "B_s", "H_w", ")", ")")
        
        expected = ('=', ('A_s',), ('mal', ('B_s',), ('H_w',)))
        result = self.parser.parse(tokens)
        
        self.assertEqual(result, expected)
    
    def testParseSimpleUnaryPredicate(self):
        tokens = ('AB', '(', 'A_w', ')')
        
        expected = (('AB', ('A_w',)), ())
        
        result = self.parser.parsePredicate(tokens)
        
        self.failUnlessEqual(result, expected)
    
    def testParseSimpleBinaryPredicate(self):
        tokens = ('=', '(', 'A_w', 'B_s', ')')
        
        expected = (('=', ('A_w',), ('B_s',)), ())
        
        result = self.parser.parsePredicate(tokens)
        
        self.failUnlessEqual(result, expected)
    
    def testParseSimpleUnaryOperator(self):
        tokens = ('qp', '(', 'A_w', ')')
        
        expected = (('qp', ('A_w',)), ())
        
        result = self.parser.parseOperator(tokens)
        
        self.failUnlessEqual(result, expected)
    
    def testParseUnaryOperatorWithConstant(self):
        tokens = ('qp', '(', '42', ')')
        
        expected = ((('qp', (42,))), ())
        
        result = self.parser.parseOperator(tokens)
        
        self.failUnlessEqual(result, expected)
    
    def testParseUnaryPredicateWithConstant(self):
        tokens = ('PZ', '(', '42', ')')
        
        expected = ((('PZ', (42,))), ())
        
        result = self.parser.parsePredicate(tokens)
        
        self.failUnlessEqual(result, expected)
    
    def testParseSimpleBinaryOperator(self):
        tokens = ('mal', '(', 'A_w', 'B_s', ')')
        
        expected = (('mal', ('A_w',), ('B_s',)), ())
        
        result = self.parser.parseOperator(tokens)
        
        self.failUnlessEqual(result, expected)
    
    def testParseBinaryOperatorWithNumericConstant(self):
        tokens = ('mal', '(', '42', '3', ')')
        
        expected = (('mal', (42,), (3,)), ())
        
        result = self.parser.parseOperator(tokens)
        
        self.failUnlessEqual(result, expected)
    
    def testParseBinaryPredicateWithNumericConstant(self):
        tokens = ('V', '(', '42', '3', ')')
        
        expected = (('V', (42,), (3,)), ())
        
        result = self.parser.parsePredicate(tokens)
        
        self.failUnlessEqual(result, expected)
    
    def testParseNestedPredicates(self):
        """This test for parsing predicates is a bit more complex."""
        tokens = ("=", "(", "A_s", "mal", "(", "B_s", "H_w", ")", ")")
        
        expected = ('=', ('A_s',), ('mal', ('B_s',), ('H_w',)))
        
        result = self.parser.parsePredicate(tokens)
        
        self.failUnlessEqual(result, (expected, ()))
    
    def testParseNestedOperators(self):
        """This test for parsing predicates is a bit more complex."""
        tokens = ('qs', '(', 'mal', '(', 'mal', '(', 'A_s', 'A_w', ')', \
            'qs', '(', 'B_s', ')', ')', ')')
        
        expected = ('qs',
            ('mal', ('mal', ('A_s',), ('A_w',)), ('qs',  ('B_s',))))
        
        result = self.parser.parseOperator(tokens)
        
        self.failUnlessEqual(result, (expected, ()))
    

class FactoryTestCase(test_utils.TestCase):
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.loader = sl.Loader()
        self.factory = sl.FieldFactory()
        self.field = sevendown.graph.Field(5, 5)
    
    def testAddNumbers(self):
        pass
        #TODO
    
    def testAddIntermediateNumbers(self):
        """Test the addIntermediateNumbers() method.
        
        It should add numbers with the given names to the given field.
        
        """
        self.assertEqual(len(self.field.cells), 25)
        self.assertEqual(len(self.field.intermediate_numbers), 0)
        
        names = ['X_0', 'X_1']
        self.factory.newCellNames = names
        self.factory.addIntermediateNumbers(self.field)
        
        self.assertEqual(len(self.field.cells), 27)
        self.assertEqual(self.field.cells[-2].name, names[0])
        self.assertEqual(self.field.cells[-1].name, names[1])
        self.assertEqual(len(self.field.intermediate_numbers), 2)
        self.assertEqual(self.field.intermediate_numbers[0].name, names[0])
        self.assertEqual(self.field.intermediate_numbers[1].name, names[1])
    
    def testGetNumberLength(self):
        """Tests the getNumberLength() method."""
        
        self.field.schema = self.factory.calculateSchema(self.field, {},[],[])
        
        result = self.factory.getNumberLength(0, 0, True, self.field)
        
        self.failUnlessEqual(result, 5)
        
        result = self.factory.getNumberLength(0, 0, False, self.field)
        
        self.failUnlessEqual(result, 5)
        
        result = self.factory.getNumberLength(1, 1, True, self.field)
        
        self.failUnlessEqual(result, 4)
        
        result = self.factory.getNumberLength(1, 1, False, self.field)
        
        self.failUnlessEqual(result, 4)
    
    def testCreateFieldWithDefinedAs(self):
        """Test that createField works correctly -- with defined as rule
        
        """
        
        field = self.factory.createField(2, 3,
            {'A': ('A', 0, 0), 'B': ('B', 1, 0)},
            (),
            ('A', 'B'),
            ( ('=', ('A_s',), ('plus', ('B_s',), (42,) )), ))
        
        # test numbers
        self.failUnlessEqual(len(field.numbers), 2, "Wrong number of Numbers generated")
        
        # one intermediate numbers should be added
        self.assertEqual(len(field.intermediate_numbers), 0, "Wrong number of intermediate numbers")
        
        # a final number of 5 rules is expected
        self.assertEqual(len(field.rules), 5, "Wrong number of rules: %s" % str(map (lambda r: str(r), field.rules)))
    
    def testCreateFieldSimple(self):
        """Test that createField works correctly -- without any nested rules.
        
        This test is not very atomic but oh my...
        
        """
        field = self.factory.createField(2, 3,
            {'A': ('A', 0, 0), 'B': ('B', 1, 0)},
            (),
            ('A', 'B'),
            ( ('=', ('A_s',), ('B_s',)), ))
        
        self.failUnlessEqual(field.height, 2)
        self.failUnlessEqual(field.width, 3)
        
        # test numbers
        self.failUnlessEqual(len(field.numbers), 2)
        
        # no intermediate numbers should be added
        self.assertEqual(len(field.intermediate_numbers), 0)
        
        # test created rules
    
    def testCreateRuleFromParseTreeWithUnaryPredicate(self):
        """Tests the createRuleFromParseTree method with an unary predicate.
        
        """
        field = sevendown.graph.Field(10, 10)
        number = sevendown.graph.Number('A_w', 1)
        field.numbers.append(number)
        parse_tree = ('PZ', ('A_w',))
        
        expected = sevendown.rules.PrimeRule(number)
        result = self.factory.createRuleFromParseTree(parse_tree, field)
        
        self.assertEqual(result.arg, expected.arg)
    
    def testCreateRuleFromParseTreeWithBinaryPredicate(self):
        """Tests the createRuleFromParseTree method with a binary predicate.
        
        """
        field = sevendown.graph.Field(10, 10)
        number1 = sevendown.graph.Number('A_w', 1)
        field.numbers.append(number1)
        number2 = sevendown.graph.Number('B_w', 1)
        field.numbers.append(number2)
        parse_tree = ('=', ('A_w',), ('B_w',))
        
        expected = sevendown.rules.EqualityRule(number1, number2)
        result = self.factory.createRuleFromParseTree(parse_tree, field)
        
        self.assertEqual(result.arg1, expected.arg1)
        self.assertEqual(result.arg2, expected.arg2)
    
    def testCreateRuleFromParseTreeWithDefinePredicate(self):
        """Tests the createRuleFromParseTree method with a := predicate."""
        field = sevendown.graph.Field(10, 10)
        number1 = sevendown.graph.Number('A_w', 1)
        field.numbers.append(number1)
        number2 = sevendown.graph.Number('B_w', 1)
        field.numbers.append(number2)
        number3 = sevendown.graph.Number('C_w', 1)
        field.numbers.append(number3)
        parse_tree = (':=', ('A_w',), ('mal', ('B_w',), ('C_w',)))
        
        expected = sevendown.rules.MultOperatorRule(number1, number2, number3)
        result = self.factory.createRuleFromParseTree(parse_tree, field)
        
        self.assertSetEqual(result.cells, expected.cells)
    
    def testFlattenParseTree_0(self):
        """Test flattenParseTree() with very simple input (identity test)."""
        parse_tree = ('AB', ('A_w',))
        expected_forest = [('AB', ('A_w',))]
        expected_cell_names = []
        
        result = self.factory.flattenParseTree(parse_tree)
        
        self.assertSetEqual(result, expected_forest)
        self.assertSetEqual(self.factory.newCellNames, expected_cell_names)
    
    def testFlattenParseTree_1(self):
        """Test flattenParseTree() with simple input (one recursion)."""
        parse_tree = ('PZ', ( 'r',  ('A_s',)))
        expected_forest = [('PZ', ('X_0',)), (':=', ('X_0',), ('r', ('A_s',)))]
        expected_cell_names = ['X_0']
        
        result = self.factory.flattenParseTree(parse_tree)
        
        self.assertSetEqual(result, expected_forest)
        self.assertSetEqual(self.factory.newCellNames, expected_cell_names)
    
    def testFlattenParseTree_2(self):
        """Test flattenParseTree() with realistic input (one recursion, 2 operands)."""
        parse_tree = ('V', ('mal', ('A_s',), ('B_w',)), ('plus', ('C_w',), ('D_s',)))
        expected_forest = [('V', ('X_0',), ('X_1',)), (':=', ('X_0',), ('mal', ('A_s',), ('B_w',))), (':=', ('X_1',), ('plus', ('C_w',), ('D_s',)))]
        expected_cell_names = ['X_0', 'X_1']
        
        result = self.factory.flattenParseTree(parse_tree)
        
        self.assertSetEqual(result, expected_forest)
        self.assertSetEqual(self.factory.newCellNames, expected_cell_names)
    
    def testFlattenParseTree_3(self):
        """Test flattenParseTree() with realistic input (two recursion, 2 operands)."""
        parse_tree = ('V', ('mal', ('A_s',), ('B_w',)), ('plus', ('C_w',), ('plus', ('D_s',), ('E_s',))))
        expected_forest = [('V', ('X_0',), ('X_1',)),  (':=', ('X_2',), ('plus', ('D_s',), ('E_s',))), (':=', ('X_0',), ('mal', ('A_s',), ('B_w',))), (':=', ('X_1',), ('plus', ('C_w',), ('X_2',))),]
        expected_cell_names = ['X_0', 'X_1', 'X_2']
        
        result = self.factory.flattenParseTree(parse_tree)
        
        self.assertSetEqual(result, expected_forest)
        self.assertSetEqual(self.factory.newCellNames, expected_cell_names)
    
    def testFlattenParseTree_4(self):
        """Test flattenParseTree() with realistic input, demonstrating shortcut semantics."""
        parse_tree = ('=', ('A_s',), ('plus', ('C_w',), ('plus', ('D_s',), ('E_s',))))
        expected_forest = [ (':=', ('X_0',), ('plus', ('D_s',), ('E_s',))), (':=', ('A_s',), ('plus', ('C_w',), ('X_0',))),]
        expected_cell_names = ['X_0',]
    	
    	result = self.factory.flattenParseTree(parse_tree)
        
        self.assertSetEqual(result, expected_forest)
        self.assertSetEqual(self.factory.newCellNames, expected_cell_names)
    
    def testFlattenParseTreeMultipleApplication(self):
        """Multiple calls to flattenParseTree must create unique cell ids."""
        parse_tree = ('V', ('mal', ('A_s',), ('B_w',)), ('plus', ('C_w',), 
            ('plus', ('D_s',), ('E_s',))))
        expected_forest = [('V', ('X_0',), ('X_1',)),  (':=', ('X_2',), 
            ('plus', ('D_s',), ('E_s',))), (':=', ('X_0',), ('mal', ('A_s',), 
            ('B_w',))), (':=', ('X_1',), ('plus', ('C_w',), ('X_2',))),]
        expected_cell_names = ['X_0', 'X_1', 'X_2']
        
        result = self.factory.flattenParseTree(parse_tree)
        
        self.assertSetEqual(result, expected_forest)
        self.assertSetEqual(self.factory.newCellNames, expected_cell_names)
    
    def testCreateRuleFromParseTreeWithConstantNumbersOnBinaryPred(self):
        tokens = ('V', (42,), (13,))
        expected = sr.MultipleOfRule(sg.Fixed(42), sg.Fixed(13))
        result = self.factory.createRuleFromParseTree(tokens, self.field)
        self.assertEqual(str(result), str(expected))
    
    def testCreateRuleFromParseTreeWithConstantNumbersOnUnaryPred(self):
        tokens = ('PZ', (42,))
        expected = sr.PrimeRule(sg.Fixed(42))
        result = self.factory.createRuleFromParseTree(tokens, self.field)
        self.assertEqual(str(result), str(expected))
    

class SpecialCaseRuleFactoryTestCase(test_utils.TestCase):
    """Tests for the class SpecialCaseRuleFactory."""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.factory = sl.SpecialCaseRuleFactory()
    
    # def testEmptyList(self):
    #     self.assertEqual(self.factory.rulesForNumber(None), [])
    

# Execute unit tests when called standalone.
if __name__ == "__main__":
    unittest.main()

# vim:sw=4:ts=4:sts=4:expandtab
