#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import unittest

import test_utils
import sevendown.rules as sr
import sevendown.graph as sg
import sevendown.config as sc
import sevendown.undo as undo

class NoLeadingZeroRuleTestCase(test_utils.TestCase):
    """Test the NoLeadingZeroRule."""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1, 1)
    
    def testRegisterDoesNoRegistering(self):
        """Test that the NoLeadingZeroRule does not register itself."""
        number = sg.Number('A_w', 3)
        number.field = self.field
        rule = sr.NoLeadingZeroRule(number)
        
        self.assertEqual(number.related_rules, [])
    
    def testStringConversion(self):
        """Test that NoLeadingZeroRule objects are properly converted into
        strings.
        
        """
        number = sg.Number('A_w', 3)
        number.field = self.field
        rule = sr.NoLeadingZeroRule(number)
        self.assertEqual(str(rule), 'NLZ(A_w)')
    
    def testApplyRemovesZeroFromFirstDigitIfNecessary(self):
        """Test remove first digit implementation."""
        digit1 = sg.Digit()
        digit1.field = self.field
        digit2 = sg.Digit()
        digit2.field = self.field
        number = sg.Number("test", 2)
        number.digits = [digit1, digit2]
        
        rule = sr.NoLeadingZeroRule(number)
        self.field.registerRule(rule)
        
        undo.startRevision()
        
        self.assertEqual(rule.apply(), [digit1])
        self.assertEqual(digit1.getFeasibleValues(), set(range(1,10)))
        
        # test that things can be rolled back
        undo.revert()
        
        self.assertEqual(digit1.getFeasibleValues(), set(range(10)))
        self.assertEqual(digit2.getFeasibleValues(), set(range(10)))
    
    def testApplyIfRemovalNotNecessary(self):
        """Test implementation if removal is necessary."""
        
        digit1 = sg.Digit()
        digit1.setFeasibleValues(set(range(1,10)))
        digit2 = sg.Digit()
        number = sg.Number("test", 2)
        number.digits = [digit1, digit2]
        
        rule = sr.NoLeadingZeroRule(number)
        self.field.registerRule(rule)
        
        undo.startRevision()
        
        self.assertEqual(rule.apply(), [])
        
        # test that things can be rolled back
        undo.revert()
        
        self.assertEqual(digit1.getFeasibleValues(), set(range(1,10)))
        self.assertEqual(digit2.getFeasibleValues(), set(range(10)))
    

class BinaryRuleTestCase(test_utils.TestCase):
    """Tests the BinaryRule class."""
    
    def testInitialization(self):
        """Test that BinaryRule objects are initialized correctly."""
        number = sg.Number('A_w', 3)
        number2 = sg.Number('B_w', 3)
        
        rule = sr.BinaryRule(number, number2)
        
        self.assertEqual(rule.arg1, number)
        self.assertEqual(rule.arg2, number2)
        self.assertEqual(rule.name, 'UNKNOWN')
    
    def testStringConversion(self):
        """Test that BinaryRule objects are properly converted to strings."""
        number = sg.Number('A_w', 3)
        number2 = sg.Number('B_w', 3)
        
        rule = sr.BinaryRule(number, number2)
        
        self.assertEqual(str(rule), 'UNKNOWN(A_w, B_w)')
    

class UnaryRuleTestCase(test_utils.TestCase):
    """Tests the UnaryRule class."""
    
    def testInitialization(self):
        """Test that UnaryRule objects are initialized correctly."""
        number = sg.Number('A_w', 3)
        
        rule = sr.UnaryRule(number)
        
        self.assertEqual(rule.arg, number)
        self.assertEqual(rule.name, 'UNKNOWN')
    
    def testStringConversion(self):
        """Test that UnaryRule objects are properly converted to strings."""
        number = sg.Number('A_w', 3)
        
        rule = sr.UnaryRule(number)
        
        self.assertEqual(str(rule), 'UNKNOWN(A_w)')
    

class DescendingRuleTestCase(test_utils.TestCase):
    """Test the DescendingRule class."""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1,1)
    
    def testStringConversion(self):
        """
        Test that DescencingRule objects are properly converted into strings.
        
        """
        number = sg.Number('A_w', 3)
        number.digits = [sg.Digit(), sg.Digit()]
        
        rule = sr.DescendingRule(number)
        self.assertEqual(str(rule), 'AB(A_w)')
    
    def testRegistration(self):
        """Test that the rule registers itself properly with the numbers."""
        number = sg.Number('A_w', 3)
        number.field = self.field
        number.digits = [sg.Digit(), sg.Digit()]
        
        rule = sr.DescendingRule(number)
        
        self.assertEqual(number.related_rules, [rule])
        self.assertEqual(number.digits[0].related_rules, [rule])
        self.assertEqual(number.digits[1].related_rules, [rule])
    
    def testApplyOnNumberWithAlreadyDescDigitsDoesNothing(self):
        """ "apply()" on a number and nothing should happen"""
        digit1 = sg.Digit()
        digit1.setFeasibleValues(set([9]))
        digit1.field = self.field
        digit2 = sg.Digit()
        digit2.setFeasibleValues(set([8]))
        digit2.field = self.field
        
        number = sg.Number('A_w', 2)
        number.field = self.field
        number.digits = [digit1, digit2]
        
        rule = sr.DescendingRule(number)
        self.field.registerRule(rule)
        
        self.assertSetEqual(rule.apply(), [])
        
        self.assertEqual(undo.revisionNo(), 1)
    
    def testApplyOnNumberWithAlreadyDescDigitsDoesNothing(self):
        """
        Make sure that the rule correctly does not remov the possible digits.
        
        """
        digit1 = sg.Digit()
        digit1.setFeasibleValues(set(range(0,10)))
        digit1.field = self.field
        digit2 = sg.Digit()
        digit2.setFeasibleValues(set(range(0,10)))
        digit2.field = self.field
        digit3 = sg.Digit()
        digit3.setFeasibleValues(set(range(0,10)))
        digit3.setFeasibleValues(set(range(0,10)))
        digit4 = sg.Digit()
        digit4.field = self.field
        digit4.field = self.field
        
        number = sg.Number('A_w', 4)
        number.field = self.field
        number.digits = [digit1, digit2, digit3, digit4]
        
        rule = sr.DescendingRule(number)
        self.field.registerRule(rule)
        
        self.assertSetEqual(rule.apply(), [])
        self.assertEqual(digit1.getFeasibleValues(), set(range(0,10)))
        self.assertEqual(digit2.getFeasibleValues(), set(range(0,10)))
        self.assertEqual(digit3.getFeasibleValues(), set(range(0,10)))
        self.assertEqual(digit4.getFeasibleValues(), set(range(0,10)))
        
        self.assertEqual(undo.revisionNo(), 1)
    
    def testApplyOnNumberWithNondescDigitsWorksCorrectly2(self):
        """Make sure that the rule correctly removes the impossible digits."""
        digit1 = sg.Digit()
        digit1.setFeasibleValues(set(range(4,10)))
        digit1.field = self.field
        digit2 = sg.Digit()
        digit2.setFeasibleValues(set(range(0,7)))
        digit2.field = self.field
        digit3 = sg.Digit()
        digit3.setFeasibleValues(set(range(1,10)))
        digit3.field = self.field
        
        number = sg.Number('A_w', 3)
        number.field = self.field
        number.digits = [digit1, digit2, digit3]
        
        rule = sr.DescendingRule(number)
        self.field.registerRule(rule)
        
        self.assertSetEqual(rule.apply(), [digit2, digit3, number])
        self.assertEqual(digit1.getFeasibleValues(), set(range(4,10)))
        self.assertEqual(digit2.getFeasibleValues(), set(range(1,7)))
        self.assertEqual(digit3.getFeasibleValues(), set(range(1,7)))
        
        self.assertEqual(undo.revisionNo(), 1)
        
        # make sure things can be rolled back
        undo.revert()
        
        self.assertEqual(digit1.getFeasibleValues(), set(range(4,10)))
        self.assertEqual(digit2.getFeasibleValues(), set(range(0,7)))
        self.assertEqual(digit3.getFeasibleValues(), set(range(1,10)))
    
    def testApplyOnDigitlessNumber(self):
        """Make sure that the rule correctly removes the impossible numbers."""
        number = sg.Number('X_0', 5)
        number.field = self.field
        
        rule = sr.DescendingRule(number)
        self.field.registerRule(rule)
        
        self.assertSetEqual(rule.apply(), [])
        
        
        number.setFeasibleValues([12345,54321,11111,12321,98765])
        self.assertSetEqual(rule.apply(), [number])

        self.assertSetEqual(number.getFeasibleValues(), [54321,11111,98765])
    

class PalindromeRuleTestCase(test_utils.TestCase):
    """Test the PalindromeRule Class."""
    
    def testStringConversion(self):
        """Test that Palindrome objects are properly converted to strings."""
        number = sg.Number('A_w', 3)
        rule = sr.PalindromeRule(number)
        self.assertEqual(str(rule), 'PAL(A_w)')
    
    def testNormalize(self):
        """
        Test that PalindromeRule creates EqualityRules on the numbers' Digits
        on normalize()
        
        """
        number = sg.Fixed(123)
        
        rule = sr.PalindromeRule(number)
        
        result = rule.normalize()
        
        self.assertEqual(len(result), 1)
        self.assert_(isinstance(result[0], sr.EqualityRule))
        self.assertEqual(result[0].arg1, number.digits[0])
        self.assertEqual(result[0].arg2, number.digits[-1])
    
    def testRegisterDoesNotRegisterWithArgument(self):
        """
        Test that PalindromeRule's register() does not register with
        argument.
        
        """
        number = sg.Fixed(123)
        
        rule = sr.PalindromeRule(number)
        
        self.assertEqual(number.getRelatedRules(), [])
    
    def testApplyOnNumberLength2(self):
        """
        Tests a simple apply on a 2 digit number and assert that the
        correct rules are created.
        
        """
        field = sg.Field(2,2)
        
        number = sg.Number('A_w', 2)
        digit1 = sg.Digit()
        digit2 = sg.Digit()
        number.digits = [digit1, digit2]
        
        rule = sr.PalindromeRule(number)
        
        new_rules = rule.normalize()
        
        self.assertEqual(len(new_rules), 1)
        
        self.assert_(isinstance(new_rules[0], sr.EqualityRule))
        self.assertEqual(new_rules[0].arg1, digit1)
        self.assertEqual(new_rules[0].arg2, digit2)
        
        self.assert_(new_rules[0] in digit1.getRelatedRules())
        self.assert_(new_rules[0] in digit2.getRelatedRules())
    
    def testApplyOnNumberLength3(self):
        """
        Tests a simple apply on a 3 digit number and assert that the
        correct rules are created.
        
        """
        field = sg.Field(2,3)
        
        number = sg.Number('A_w', 3)
        digit1 = sg.Digit()
        digit2 = sg.Digit()
        digit3 = sg.Digit()
        number.digits = [digit1, digit2, digit3]
        
        rule = sr.PalindromeRule(number)
        
        new_rules = rule.normalize()
        
        self.assertEqual(len(new_rules), 1)
        
        self.assert_(isinstance(new_rules[0], sr.EqualityRule))
        self.assertEqual(new_rules[0].arg1, digit1)
        self.assertEqual(new_rules[0].arg2, digit3)
        
        self.assert_(new_rules[0] in digit1.getRelatedRules())
        self.assert_(new_rules[0] not in digit2.getRelatedRules())
        self.assert_(new_rules[0] in digit3.getRelatedRules())
    

class RemoveValuesFromDigitsRule(test_utils.TestCase):
    """Test case for the RemoveValuesFromDigitsRule."""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1,1)
    
    def testStringConversion(self):
        digit = sg.Digit('(0,0)')
        digit.field = self.field
        
        rule = sr.RemoveValuesFromDigitsRule(digit, [1,9])
        
        self.assertEqual(str(rule), 'REMOVE[1,9](%s)' % str(digit))
    
    def testRuleDoesNotRegisterItself(self):
        """Test that the rule does not register itself with the argument."""
        digit = sg.Digit(coords=(0,0))
        digit.field = self.field
        
        rule = sr.RemoveValuesFromDigitsRule(digit, [0])
        
        self.assertEqual(digit.getRelatedRules(), [])
    
    def testApplyWithChanges(self):
        digit = sg.Digit(coords=(0,0))
        digit.field = self.field
        
        rule = sr.RemoveValuesFromDigitsRule(digit, [1,9])
        
        self.assertEqual(rule.apply(), [digit])
        self.assertEqual(digit.getFeasibleValues(), set([0,2,3,4,5,6,7,8]))
        
        # test that things can be rolled back
        
        undo.revert()
        
        self.assertEqual(digit.getFeasibleValues(), set(range(10)))
    

class PrimeRuleTestCase(test_utils.TestCase):
    """Test the PrimeRule Class."""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1,1)
    
    def testStringConversion(self):
        """Test that PrimeRule objects are properly converted to strings."""
        number = sg.Number('A_w', 3)
        rule = sr.PrimeRule(number)
        self.assertEqual(str(rule), 'PZ(A_w)')
    
    def testNormalizeWorksCorrectlyOnNumber(self):
        """
        Test that normalize() creates a RemoveValuesFromDigitsRule if the
        argument is a Number.
        
        """
        number = sg.Number('A_w', 2)
        number.digits = [sg.Digit(), sg.Digit()]
        
        rule = sr.PrimeRule(number)
        
        normalized = rule.normalize()
        self.assertEqual(len(normalized), 2)
        self.assertEqual(normalized[0], rule)
        self.assert_(isinstance(normalized[1], sr.RemoveValuesFromDigitsRule))
        self.assertEqual(normalized[1].arg, rule.arg.digits[-1])
    
    def testNormalizeWorksCorrectlyOnFixed(self):
        """
        Normalize should create a RemoveValuesFromDigitsRule on last digit.
        
        """
        fixed = sg.Fixed(13)
        rule = sr.PrimeRule(fixed)
        
        result = rule.normalize()
        
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], rule)
        self.assert_(isinstance(result[1], sr.RemoveValuesFromDigitsRule))
        self.assertEqual(result[1].arg, fixed.digits[-1])
    
    def testApplyDoesNothingOnLargeNumber(self):
        number = sg.Number('A_w', 8)
        number.digits = [sg.Digit()]*8
        
        number.field = self.field
        for d in number.digits:
            d.field = self.field
        
        rule = sr.PrimeRule(number)
        
        self.assertEqual(rule.apply(), [])
        self.assertEqual(number.digits[0].getFeasibleValues(), set(range(10)))
    
    def testApplyDoesSomethingOnMediumLargeNumber(self):
        number = sg.Number('A_w', 4)
        number.digits = [sg.Digit()]*4
        
        number.field = self.field
        for d in number.digits:
            d.field = self.field
        
        rule = sr.PrimeRule(number)
        
        self.assertEqual(rule.apply(), [number])
        self.assertEqual(len(number.getFeasibleValues()), 1061)

    def testApplySetsNonPrimeDoneNumberToEmpty(self):
        number = sg.Number('A_w', 2)
        number.digits = [sg.Digit(), sg.Digit()]
        
        number.field = self.field
        for d in number.digits:
            d.field = self.field
        
        number.force(set([12]))
        
        rule = sr.PrimeRule(number)
        
        self.assertEqual(rule.apply(), [number])
        self.assertEqual(number.getFeasibleValues(), set())
    
    def testApplyReturnsEmptyListOnDonePrime(self):
        number = sg.Number('A_w', 2)
        number.digits = [sg.Digit(), sg.Digit()]
        
        number.field = self.field
        for d in number.digits:
            d.field = self.field
        
        number.force(set([11]))
        
        rule = sr.PrimeRule(number)
        
        self.assertEqual(rule.apply(), [])
        self.assertEqual(number.getFeasibleValues(), set([11]))
  
    def testApplySetsBigNonPrimeDoneNumberToEmpty(self):
        number = sg.Number('A_w')
        
        number.field = self.field
        
        number.force(set([97026267]))
        
        rule = sr.PrimeRule(number)
        
        self.assertEqual(rule.apply(), [number])
        self.assertEqual(number.getFeasibleValues(), set())
    
    def testApplyReturnsEmptyListOnDoneBigPrime(self):
        number = sg.Number('A_w')
        
        number.field = self.field
        
        number.force(set([97026263]))
        
        rule = sr.PrimeRule(number)
        
        self.assert_(rule.is_prime(97026263))
        self.assertEqual(rule.apply(), [])
        self.assertEqual(number.getFeasibleValues(), set([97026263]))
    

class SquareTestCase(test_utils.TestCase):
    """Tests the Square Class"""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(2,2)
    
    def testStringConversion(self):
        """Test that SquareRule objects are properly converted to strings."""
        number = sg.Number('A_w', 3)
        rule = sr.SquareRule(number)
        self.assertEqual(str(rule), 'Q(A_w)')
    
    def testSquareRuleApplyLengthSmallAction(self):
        """ Test if SquareRule section small length of number works."""
        number = sg.Number('A_w', 3)
        number.digits = [sg.Digit()] * 3
        number.force(set(range(100,1000)))
        rule = sr.SquareRule(number)
        self.assertEqual(str(rule), 'Q(A_w)')
        
        self.assertSetNotEqual(rule.apply(), [])

    def testSquareRuleApplyiCardSmallAction(self):
        """ Test if SquareRule section small length of number works."""
        number = sg.Number('A_w', 3)
        number.digits = [sg.Digit()] * 3
        number.force(set(range(100,1000)))
        rule = sr.SquareRule(number)
        self.assertEqual(str(rule), 'Q(A_w)')
        
        self.assertSetNotEqual(rule.apply(), [])
    
    def testApplyOn625(self):
        number = sg.Number("A_w")
        number.force(set([625]))
        number.field = self.field
        
        rule = sr.SquareRule(number)
        
        result = rule.apply()
        
        self.assertEqual([], result)
        self.assertEqual(set([625]), number.getFeasibleValues())
    

class UniqueDigitsRuleTestCase(test_utils.TestCase):
    """Test the UniqueDigitsRule Class."""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1,1)
    
    def testStringConversion(self):
        """Test that UniqueDigitsRule objects are properly converted to
        strings.
        
        """
        number = sg.Number('A_w', 3)
        number.digits = []
        rule = sr.UniqueDigitsRule(number)
        self.assertEqual(str(rule), 'U(A_w)')
    

class UniqueDigitsApplicationRuleTestCase(test_utils.TestCase):
    """Test UniqueDigitsRule.appy() with various parameters."""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1,1)
    
    def testApplyOnLargeNumberWithNoFixedDigitDoesNothing(self):
        """apply() on a number which does not have an isDone() Digit."""
        number = sg.Number('A_w', 5)
        digit1 = sg.Digit()
        digit1.field = self.field
        digit1.force(set([0, 1]))
        number.digits[0] = digit1
        digit2 = sg.Digit()
        digit2.field = self.field
        digit2.force(set([0, 1]))
        number.digits[1] = digit1
        number.digits[2] = sg.Digit()
        number.digits[3] = sg.Digit()
        number.digits[4] = sg.Digit()
        
        rule = sr.UniqueDigitsRule(number)
        self.field.registerRule(rule)
        
        self.assertEqual(rule.apply(), [])
        
        self.assertEqual(undo.revisionNo(), 1)
    
    def testApplyOnNumberWithOneFixedDigitRemovesItFromOther(self):
        """apply() on a number which does not have an isDone() Digit."""
        number = sg.Number('A_w', 2)
        digit1 = sg.Digit()
        digit1.field = self.field
        digit1.force(set([0, 1]))
        number.digits[0] = digit1
        digit2 = sg.Digit()
        digit2.field = self.field
        digit2.force(set([0]))
        number.digits[1] = digit2
        
        rule = sr.UniqueDigitsRule(number)
        self.field.registerRule(rule)
        
        undo.startRevision()
        
        self.assertEqual(rule.apply(), [digit1])
        
        self.assertEqual(digit1.getFeasibleValues(), set([1]))
        self.assertEqual(digit2.getFeasibleValues(), set([0]))
        
    
    def testApplyChainReaction(self):
        """Make sure that more than one digit is changed."""
        number = sg.Number('A_w', 3)
        digit1 = sg.Digit()
        digit1.field = self.field
        digit1.force(set([0]))
        digit2 = sg.Digit()
        digit2.field = self.field
        digit2.force(set([0,1]))
        digit3 = sg.Digit()
        digit3.field = self.field
        digit3.force(set([0,1,2]))
        
        number.digits = (digit1, digit2, digit3)
        
        rule = sr.UniqueDigitsRule(number)
        self.field.registerRule(rule)
        
        self.assertSetEqual(rule.apply(), [digit2, digit3])
        self.assertEqual(digit1.getFeasibleValues(), set([0]))
        self.assertEqual(digit2.getFeasibleValues(), set([1]))
        self.assertEqual(digit3.getFeasibleValues(), set([2]))
    
    def testApplyMakesDigitEmptyRegresseion(self):
        """
        Test that the following does not happen any more: UniqueDigitsRule
        is applied to a number.  In this course, a digit is first marked as
        done and then marked as empty.
        
        """
        
        number = sg.Number('A_w', 3)
        number.digits = [sg.Digit(), sg.Digit(), sg.Digit()]
        for d in number.digits: d.field = self.field
        
        number.digits[0].force(set([0]))
        number.digits[1].force(set([0,1]))
        number.digits[2].force(set([1]))
        
        rule = sr.UniqueDigitsRule(number)
        self.field.registerRule(rule)
        
        result = rule.apply()
        
        self.assert_(len(result) > 0)
        self.assert_(number.digits[0].isEmpty() or number.digits[1].isEmpty() or number.digits[2].isEmpty())
    
    def testApplyMakesDigitEmptyRegresseion2(self):
        """
        Test that the following works
        """
        
        number = sg.Number('A_w', 5)
        number.digits = [sg.Digit(), sg.Digit(), sg.Digit(), sg.Digit(), sg.Digit()]
        for d in number.digits: d.field = self.field
        
        for d, v in zip(number.digits, [8,1,1,2,3]):
            d.force(set([v]))

        rule = sr.UniqueDigitsRule(number)
        self.field.registerRule(rule)
        
        result = rule.apply()
        
        self.assert_(len(result) > 0)
        self.assert_(number.digits[0].isEmpty())
    

class CrossSumRuleTestCase(test_utils.TestCase):
    """Tests the CrossSumeRule class."""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1,1)
    
    def testFirstArgumentGetsIntersectedCorrectlyIfSecondIsDone(self):
        number1 = sg.Number("B_w", length=2)
        number2 = sg.Number("A_w", length=2)
        number2.digits = [sg.Digit(), sg.Digit()]
        
        rule = sr.CrossSumRule(number1, number2)
        
        number1.field = self.field
        number2.field = self.field
        for d in number2.digits:
            d.field = self.field
        
        number2.digits[0].force(set([6]))
        number2.digits[1].force(set([6]))
        number1.force(set([10, 12]))
        
        undo.startRevision()
        
        result = rule.apply()
        
        self.assertSetEqual(result, [number1])
        self.assertEqual(number1.getFeasibleValues(), set([12]))
        
        # make sure things can be rolled back
        undo.revert()
        
        self.assertEqual(number1.getFeasibleValues(), set([12,10]))
    
    def testRegisterFixedNumber(self):
        """Make sure CrossSumRules register themselves correctly."""
        fixed = sg.Fixed(7)
        number = sg.Number("A_w", length=2)
        number.digits = [sg.Digit(), sg.Digit()]
        
        rule = sr.CrossSumRule(fixed, number)
        
        self.assert_(rule in fixed.getRelatedRules())
        self.assert_(rule in number.getRelatedRules())
        self.assert_(rule in number.digits[0].getRelatedRules())
        self.assert_(rule in number.digits[1].getRelatedRules())
    
    def testRegisterNumberNumber(self):
        """Make sure CrossSumRules register themselves correctly."""
        number1 = sg.Number("A_w", length=2)
        number1.digits = [sg.Digit(), sg.Digit()]
        number2 = sg.Number("A_w", length=2)
        number2.digits = [sg.Digit(), sg.Digit()]
        
        rule = sr.CrossSumRule(number1, number2)
        
        self.assert_(rule in number1.getRelatedRules())
        self.assert_(rule in number1.digits[0].getRelatedRules())
        self.assert_(rule in number1.digits[1].getRelatedRules())
        self.assert_(rule in number2.getRelatedRules())
        self.assert_(rule in number2.digits[0].getRelatedRules())
        self.assert_(rule in number2.digits[1].getRelatedRules())
    
    def testStringConversion(self):
        """
        Test that CrossSumRule objects are correctly converted to strings.
        
        """
        fixed = sg.Fixed(42)
        number = sg.Number('A_w')
        
        rule = sr.CrossSumRule(fixed, number)
        
        self.assertEqual(str(rule), 'CrossSumIs(42, %s)' % str(number))
    
    def testDoesNothingIfFirstArgumentIsNotDone(self):
        number1 = sg.Number('A_w', length=2)
        number1.digits = [sg.Digit(), sg.Digit()]
        number2 = sg.Number('A_s', length=2)
        number2.digits = [sg.Digit(), sg.Digit()]
        
        for n in [number1, number2]:
            n.field = self.field
            for d in n.digits:
                d.field = self.field
        
        rule = sr.CrossSumRule(number1, number2)
        self.field.registerRule(rule)
        
        result = rule.apply()
        self.assertEqual(result, [])
    
    def testDoesUpperAndLowerLimitationWhenFeasible(self):
        fixed = sg.Fixed(5)
        fixed.field = self.field
        number = sg.Number('A_w', length=2)
        number.digits = [sg.Digit(), sg.Digit()]
        
        number.field = self.field
        for d in number.digits:
            d.field = self.field
        
        rule = sr.CrossSumRule(fixed, number)
        self.field.registerRule(rule)
        
        result = rule.apply()
        self.assertSetEqual(result, number.digits + [number])
        
        # make sure things can be rolled back
        undo.revert()
        
        self.assertEqual(number.digits[0].getFeasibleValues(), set(range(10)))
        self.assertEqual(number.digits[1].getFeasibleValues(), set(range(10)))
    
    def testSetsSecondNumberToEmptySetIfRuleNotSatisfiable(self):
        fixed = sg.Fixed(0)
        fixed.field = self.field
        number = sg.Number('A_w', length=2)
        number.digits = [sg.Digit(), sg.Digit()]
        
        number.field = self.field
        for d in number.digits:
            d.field = self.field
        number.digits[0].force(set(range(1,10)))
        
        rule = sr.CrossSumRule(fixed, number)
        self.field.registerRule(rule)
        
        result = rule.apply()
        self.assertSetEqual(result, number.digits)
        self.assertEqual(number.digits[0].getFeasibleValues(), set())
        
        # make sure things can be rolled back
        undo.revert()
        
        self.assertEqual(number.digits[0].getFeasibleValues(), set(range(10)))
        self.assertEqual(number.digits[1].getFeasibleValues(), set(range(10)))
    
    def testUpdatesSecondNumbersPossibleValuesCorrectly(self):
        fixed = sg.Fixed(3)
        fixed.field = self.field
        number = sg.Number('A_w')
        number.digits = [sg.Digit(), sg.Digit()]
        number.digits[0].setFeasibleValues(set(range(1, 10)))
        
        number.field = self.field
        for i, d in enumerate(number.digits):
            d.field = self.field
            d.name = "%d" % i
        
        rule = sr.CrossSumRule(fixed, number)
        self.field.registerRule(rule)
        
        undo.startRevision()
        
        result = rule.apply()
        self.assertSetEqual(result, number.digits + [number])
        self.assertEqual(number.getFeasibleValues(), set([12, 21, 30]))
        self.assertEqual(number.digits[0].getFeasibleValues(), set([1,2,3]))
        self.assertEqual(number.digits[1].getFeasibleValues(), set([0,1,2]))
        
        # make sure things can be rolled back
        undo.revert()
        
        self.assertEqual(number.getFeasibleValues(), set(range(10,100)))
        self.assertEqual(number.digits[0].getFeasibleValues(), set(range(1,10)))
        self.assertEqual(number.digits[1].getFeasibleValues(), set(range(10)))
    
    def testIntersectsFirstNumberIfSecondIsDone(self):
        fixed = sg.Fixed(3)
        number = sg.Number('A_w')
        number.digits = [sg.Digit(), sg.Digit()]
        number.digits[0].setFeasibleValues(set([1]))
        number.digits[1].setFeasibleValues(set([0]))
        
        fixed.field = self.field
        number.field = self.field
        for i, d in enumerate(number.digits):
            d.field = self.field
            d.name = "%d" % i
        
        rule = sr.CrossSumRule(fixed, number)
        self.field.registerRule(rule)
        
        result = rule.apply()
        
        self.assertEqual(result, [fixed])
        self.assertEqual(set([]), fixed.getFeasibleValues())
    

class MultipleOfRuleTestCase(test_utils.TestCase):
    """Tests the MultipleOfRule class."""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1,1)
    
    def testStringConversion(self):
        """Test that MultipleOfRule objects are properly converted to
        strings.
        
        """
        number = sg.Number('A_w', 3)
        number.digits = [sg.Digit(), sg.Digit(), sg.Digit()]
        number2 = sg.Number('B_w', 3)
        number2.digits = [sg.Digit(), sg.Digit(), sg.Digit()]
        rule = sr.MultipleOfRule(number, number2)
        self.assertEqual(str(rule), 'V(A_w, B_w)')
    
    def testNormalizeReturnsSelfOnly(self):
        """normalize() should only return a list with the rule itself"""
        number1 = sg.Fixed(8)
        number2 = sg.Fixed(2)
        
        rule = sr.MultipleOfRule(number1, number2)
        
        result = rule.normalize()
        
        self.assertEqual(result, [rule])
    
    def testRegisterRegistersWithArguments(self):
        """rule should register itself with its arguments"""
        number1 = sg.Fixed(8)
        number2 = sg.Fixed(2)
        
        rule = sr.MultipleOfRule(number1, number2)
        
        self.assertEqual(number1.getRelatedRules(), [rule])
        self.assertEqual(number2.getRelatedRules(), [rule])
    
    def testApplyMultipleOfDoneValid(self):
        """MultipleOfRule if arguments are done, the rule is not satisfied."""
        number1 = sg.Fixed(8)
        number2 = sg.Fixed(2)
        
        rule = sr.MultipleOfRule(number1, number2)
        
        result = rule.apply()
        
        self.assertEqual(result, [])
        self.assertEqual(number1.value(), 8)
        self.assertSetEqual(number1.getFeasibleValues(), [8])
        self.assertEqual(number2.value(), 2)
        self.assertSetEqual(number2.getFeasibleValues(), [2])
    
    def testApplyMultipleOfDoneInvalid(self):
        """MultipleOfRule if arguments are done, the rule is satisfied."""
        number1 = sg.Fixed(8)
        number2 = sg.Fixed(3)
        
        rule = sr.MultipleOfRule(number1, number2)
        
        result = rule.apply()
        
        self.assertEqual(result, [number1, number2])
        self.assertEqual(number1.getFeasibleValues(), set())
        self.assertEqual(number2.getFeasibleValues(), set())
    
    def testApplyMultipleOfNotDoneNoDecisionPossible(self):
        """
        MultipleOfRule if arguments are not done and no decision is possible
        as of yet.
        
        """
        number1 = sg.Number('A_w', 2)
        number1.digits = [sg.Digit(), sg.Digit()]
        number2 = sg.Number('B_w', 2)
        number2.digits = [sg.Digit(), sg.Digit()]
        
        rule = sr.MultipleOfRule(number1, number2)
        
        result = rule.apply()
        
        self.assertEqual(result, [])
    
    def testApplyMultipleOfNotDoneArg2InfluencesArg1(self):
        """
        MultipleOfRule if arguments 2 influences argument 1 (i.e. sets arg1's
        valid values to multiples of arg2's valid values).
        
        """
        number1 = sg.Number('A_w', 2)
        number1.digits = [sg.Digit(), sg.Digit()]
        number2 = sg.Number('B_w', 2)
        number2.digits = [sg.Digit(), sg.Digit()]
        rule = sr.MultipleOfRule(number1, number2)
        
        for n in [number1, number2]:
            n.field = n
            for digit in n.digits:
                digit.field = self.field
        
        number2.digits[-1].force(set([2]))
        
        undo.startRevision()
        
        # apply rule and test result
        result = rule.apply()
        
        self.assertEqual(result, [number1.digits[-1]])
        self.assertEqual(number1.digits[-1].getFeasibleValues(), set([0,2,4,6,8]))
        
        # make sure things can be rolled back
        undo.revert()
        self.assertEqual(number1.digits[-1].getFeasibleValues(), set(range(10)))
    
    def testApplyMultipleOfNotDoneArg2InfluencesArg1Regression(self):
        """
        MultipleOfRule if arguments 2 influences argument 1 (i.e. sets arg1's
        valid values to multiples of arg2's valid values).
        
        Test this with "multiples of 5" so the "< instead of modulo"
        regression is tested.
        
        """
        number1 = sg.Number('A_w', 2)
        number1.digits = [sg.Digit(), sg.Digit()]
        number2 = sg.Number('B_w', 2)
        number2.digits = [sg.Digit(), sg.Digit()]
        rule = sr.MultipleOfRule(number1, number2)
        
        for n in [number1, number2]:
            n.field = n
            for digit in n.digits:
                digit.field = self.field
        
        number1.digits[-1].force(set([5]))
        number2.digits[-1].force(set([3]))
        
        undo.startRevision()
        
        # apply rule and test result
        result = rule.apply()
        
        self.assertEqual(result, [])
    
    def testApplyMultipleOfNotDoneArg1InfluencesArg2(self):
        """
        MultipleOfRule if arguments 1 influences argument 2 (i.e. sets arg1's
        valid values to the factors (mod 10) of arg2's value.
        
        """
        number1 = sg.Number('A_w', 2)
        number1.digits = [sg.Digit(), sg.Digit()]
        number2 = sg.Number('B_w', 2)
        number2.digits = [sg.Digit(), sg.Digit()]
        rule = sr.MultipleOfRule(number1, number2)
        
        for n in [number1, number2]:
            n.field = n
            for digit in n.digits:
                digit.field = self.field
        
        number1.digits[-1].force(set([5]))
        
        undo.startRevision()
        
        # apply rule and test result
        result = rule.apply()
        
        self.assertEqual(result, [number2.digits[-1]])
        self.assertEqual(number2.digits[-1].getFeasibleValues(), set([1,3,5,7,9]))
        
        # make sure things can be rolled back
        undo.revert()
        
        self.assertEqual(number2.digits[-1].getFeasibleValues(), set(range(10)))
    
    def testApplyWhenSecondIsDoneAndFirstIsNot(self):
        """
        Test the case V(X, 123) where X is 3 digits long and nothing is known
        about its digits yet.
        
        """
        # create numbers
        number = sg.Number('A_w', 3)
        number.digits = [sg.Digit(), sg.Digit(), sg.Digit()]
        fixed = sg.Fixed(123)
        
        # set field
        for d in number.digits:
            d.field = self.field
        number.field = self.field
        fixed.field = self.field
        
        rule = sr.MultipleOfRule(number, fixed)
        
        # apply rule
        result = rule.apply()
        
        self.assertEqual(set(result), set([number]))
        self.assertEqual(number.getFeasibleValues(), set(123*i for i in range(0,9)))
    

class EqualityRuleTestCaseOnDigits(test_utils.TestCase):
    """
    Test case for EqualityRule on Digit objects.
    
    """
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1,1)
    
    def testNormalize(self):
        """EqualityRule should not normalize itself."""
        digit1 = sg.Digit()
        digit1.field = self.field
        digit2 = sg.Digit()
        digit2.field = self.field
        
        rule = sr.EqualityRule(digit1, digit2)
        
        result = rule.normalize()
        expected = [rule]
        
        self.assertEqual(result, expected)
    
    def testRegister(self):
        """EqualityRule should register itself with the Digits."""
        digit1 = sg.Digit()
        digit1.field = self.field
        digit2 = sg.Digit()
        digit2.field = self.field
        
        rule = sr.EqualityRule(digit1, digit2)
        
        self.assertEqual(digit1.getRelatedRules(), [rule])
        self.assertEqual(digit2.getRelatedRules(), [rule])
    

class EqualityRuleTestCaseOnNumbers(test_utils.TestCase):
    """
    Test case for EqualityRule on Number objects.
    
    """
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1,1)
        
        self.number1 = sg.Number('A_w', 2)
        self.number1.digits = [sg.Digit(), sg.Digit()]
        self.number1.field = self.field
        self.number2 = sg.Number('B_w', 2)
        self.number2.digits = [sg.Digit(), sg.Digit()]
        self.number2.field = self.field
    
    def testNormalize(self):
        """normalize() should create EqualityRules on the digits"""
        rule = sr.EqualityRule(self.number1, self.number2)
        
        result = rule.normalize()
        
        self.assertEqual(len(result), 2)
        for i in range(len(result)):
            self.assert_(isinstance(result[i], sr.EqualityRule))
            self.assertEqual(result[i].arg1, self.number1.digits[i])
            self.assertEqual(result[i].arg2, self.number2.digits[i])
    
    def testRegister(self):
        """EqualityRule should not register itself at the numbers."""
        rule = sr.EqualityRule(self.number1, self.number2)
        
        self.assertEqual(self.number1.getRelatedRules(), [])
        self.assertEqual(self.number2.getRelatedRules(), [])
    

class NumberRuleTestCase(test_utils.TestCase):
    def setUp(self):
        test_utils.TestCase.setUp(self)
        
        self.digit1 = sg.Digit()
        self.digit2 = sg.Digit()
        self.digit3 = sg.Digit()
        self.digit4 = sg.Digit()

        self.number = sg.Number("Test", 4)
        self.number.digits = [self.digit1, self.digit2, self.digit3, self.digit4]

        self.rule = sr.NumberRule(self.number)

    def testDigitToNumber(self):
        # no explicit set yet
        ret = self.rule.apply()
        self.assertEqual(ret, [])

        self.digit2.setFeasibleValues(set((1,)))
        self.digit4.setFeasibleValues(set((9,)))
        
        ret = self.rule.apply()
        self.assertEqual(ret, [self.number])

        self.digit1.setFeasibleValues(set((1,2)))
        self.digit3.setFeasibleValues(set((1,2)))

        ret = self.rule.apply()
        self.assertEqual(ret, [self.number])
        self.assertSetEqual(self.number.getFeasibleValues(),
                [1119, 1129, 2119, 2129])

    def testNumberToDigit(self):
        # no explicit set yet
        ret = self.rule.apply()
        self.assertEqual(ret, [])
        self.digit1.force(set(range(1,10)))

        # force values
        self.number.force(set((1129,1139,1219,1119,2119,3119,4119,5119,6119,7119,8119,9119)))

        ret = self.rule.apply()
        self.assertSetEqual(ret, [self.digit2, self.digit3, self.digit4])

        self.assertSetEqual(self.digit1.getFeasibleValues(), range(1,10))
        self.assertSetEqual(self.digit2.getFeasibleValues(), [1,2])
        self.assertSetEqual(self.digit3.getFeasibleValues(), [1,2,3])
        self.assertSetEqual(self.digit4.getFeasibleValues(), [9])


class TestMultOperatorRule(test_utils.TestCase):
    """Test MultOperatorRule objects, and therefore OperatorRule objects.
    """
    
    def testApplyArg2Done(self):
        """Check MultOperatorRule behaviour when arg2 isDone and arg1 is None."""
        res = sg.Number('R')
        arg1 = sg.Number('A_1')
        arg2 = sg.Number('A_2')
        arg2.force(set((2,3)))
        rule = sr.MultOperatorRule(res, arg1, arg2)
        
        self.assert_(rule.sourceCard() >= sc.ALMOST_INFINITE)
        self.assertEqual(str(rule), "IsMal(R,A_1,A_2)")
        self.assertEqual([], rule.apply())
    
    def testApplyBothArgsDone(self):
        """Check MultOperatorRule behaviour when boths args are done"""
        res = sg.Number('R')
        arg1 = sg.Fixed(2)
        arg2 = sg.Fixed(42)
        rule = sr.MultOperatorRule(res, arg1, arg2)
        
        self.assertEqual(rule.sourceCard(), 1)
        self.assertEqual(rule.apply(), [res])
        self.assertEqual(res.value(), 2*42)
    
    def testApplyBothArgsSmall(self):
        """Check MultOperatorRule behaviour when boths args are done"""
        res = sg.Number('R')
        arg1 = sg.Number('A_1')
        arg1.force(set((10,15)))
        arg2 = sg.Number('A_2')
        arg2.force(set((2,3)))
        rule = sr.MultOperatorRule(res, arg1, arg2)
        
        self.assertEqual(rule.sourceCard(), 4)
        self.assertEqual(rule.apply(), [res])
        self.assertSetEqual(res.getFeasibleValues(), [20, 30, 45])
    
    def testExecute(self):
        """This test increases the number of tests we have"""
        res = sg.Number('R')
        arg1 = sg.Fixed(2)
        arg2 = sg.Fixed(42)
        rule = sr.MultOperatorRule(res, arg1, arg2)
        for x,y in (1,2), (3,4), (23,42), (112,110), (1024,768):
            self.assertEqual(rule.execute(x,y), x*y)


class TestFlipOperatorRule(test_utils.TestCase):
    """Test FlipOperatorRule objects.
    """

    def testBothWithDigits(self):
        """Check FlipOperatorRule with two digithaving arguments"""
        arg1 = sg.FieldNumber('A_1', 1, 1,3)
        arg1.digits = [sg.Digit(), sg.Digit(), sg.Digit()]
        arg2 = sg.FieldNumber('A_2', 1, 1,3)
        arg2.digits = [sg.Digit(), sg.Digit(), sg.Digit()]

        self.field = sg.Field(3,1)
        arg1.field = self.field
        arg2.field = self.field
        for d in arg1.digits + arg2.digits:
            d.field = self.field

        rule = sr.FlipOperatorRule(arg1, arg2)

        norm = rule.normalize()

        self.assertEqual(len(norm), 3)
        self.assert_(arg1.digits[0].related_rules == arg2.digits[2].related_rules)

    
    def noTestBothWithDigits(self):
        """Check FlipOperatorRule with digithaving argument first"""
        arg1 = sg.Number('A_1', 3)
        arg1.digits = [sg.Digit(), sg.Digit(), sg.Digit()]
        arg2 = sg.Number('A_2', 3)

        rule = sr.FlipOperatorRule(arg1, arg2)
        norm = rule.normalize()

        self.assert_(len(norm) == 1)
        self.assert_(isinstance(norm[0], sr.NumberRule))
        self.assertEqual(arg2.digits, arg1.digits[::-1])
        self.assertSetEqual(arg1.digits[0].related_rules, [norm[0]])

    def noTestBothWithDigits2(self):
        """Check FlipOperatorRule with digithaving argument second"""
        arg1 = sg.Number('A_1', 3)
        arg2 = sg.Number('A_2', 3)
        arg2.digits = [sg.Digit(), sg.Digit(), sg.Digit()]

        rule = sr.FlipOperatorRule(arg1, arg2)
        norm = rule.normalize()

        self.assert_(len(norm) == 1)
        self.assert_(isinstance(norm[0], sr.NumberRule))
        self.assertEqual(arg1.digits, arg2.digits[::-1])
        self.assertSetEqual(arg1.digits[0].related_rules, [norm[0]])


# Execute unit tests when called standalone.
if __name__ == "__main__":
    unittest.main()

# vim:sw=4:ts=4:sts=4:expandtab
