#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import unittest
import random

import sevendown.solver as ss
import sevendown.graph as sg
import sevendown.rules as sr

class TestField1(sg.Field):
    def __init__(self):
        self.transaction = sb.TransactionHelper()
        
        d1 = sg.Digit()
        d1.field = self
        d2 = sg.Digit()
        d2.field = self
        d3 = sg.Digit()
        d3.field = self
        d3.setFeasibleValues(set([1]))
        r1 = sr.EqualityRule(d1,d2)
        r2 = sr.EqualityRule(d2, d3)
        self.cells = [d1, d2, d3]
        self.rules = [r1, r2]
    

class TestField2(sg.Field):
    def __init__(self):
        self.transaction = sb.TransactionHelper()
        
        d1 = sg.Digit()
        d1.field = self
        d2 = sg.Digit()
        d2.field = self
        d2.setFeasibleValues(set([1]))
        r1 = sr.EqualityRule(d1, d2)
        self.cells = [d1]
        self.rules = [r1]
    

class TestField3(sg.Field):
    """A simple test field with an IntermediateNumber.
    
    This cannot be solved by StupidSolver.
    """
    
    def __init__(self):
        self.transaction = sb.TransactionHelper()
        
        fixed1 = sg.Fixed(1)
        fixed2 = sg.Fixed(1)
        
        i1 = sg.IntermediateNumber('X_0')
        
        r1 = sr.DefinedAsRule(i1, so.AdditionOperator(fixed1, fixed2))
        
        self.cells = [ i1 ]
        self.numbers = []
        self.intermediate_numbers = [ i1 ]
        self.rules = [ r1 ]
    


class NextSplittableSelectorTestCase(unittest.TestCase):
    """Test the Solver class."""
    
    def testSelectFirst(self):
        """
        Test that selectFirst() returns the first cells' index.
        
        """
        field = sg.Field(2,2)
        field.digits[0][0].setFeasibleValues(set(range(1,9)))
        field.digits[0][1].setFeasibleValues(set(range(4)))
        field.digits[1][0].setFeasibleValues(set())
        field.digits[1][1].setFeasibleValues(set(range(10)))
        
        result = ss.NextSplittableSelector.selectFirst(field.cells)
        
        self.assertEqual(field.cells[result], field.digits[0][0])
    
    def testSelectSmallest(self):
        """
        Test that selectSmallest() returns the first cells' index.
        
        """
        field = sg.Field(2,2)
        field.digits[0][0].setFeasibleValues(set(range(1,9)))
        field.digits[0][1].setFeasibleValues(set(range(4)))
        field.digits[1][0].setFeasibleValues(set())
        field.digits[1][1].setFeasibleValues(set(range(10)))
        
        result = ss.NextSplittableSelector.selectSmallest(field.cells)
        
        self.assertEqual(field.cells[result], field.digits[0][1])
    
    def testSelectLargest(self):
        """
        Test that selectLargest() returns the first cells' index.
        
        """
        field = sg.Field(2,2)
        field.digits[0][0].setFeasibleValues(set(range(1,9)))
        field.digits[0][1].setFeasibleValues(set(range(4)))
        field.digits[1][0].setFeasibleValues(set())
        field.digits[1][1].setFeasibleValues(set(range(10)))
        
        result = ss.NextSplittableSelector.selectLargest(field.cells)
        
        self.assertEqual(field.cells[result], field.digits[1][1])
    
    def testSelectRandom(self):
        """
        Test that selectRandom() returns the correct "random" cell.
        
        """
        field = sg.Field(2,2)
        field.digits[0][0].setFeasibleValues(set(range(1,9)))
        field.digits[0][1].setFeasibleValues(set(range(4)))
        field.digits[1][0].setFeasibleValues(set())
        field.digits[1][1].setFeasibleValues(set(range(10)))
        
        random.seed(42)
        expected = random.randrange(len(field.cells))
        while not field.cells[expected].isSplittable():
            expected = random.randrange(len(field.cells))
        
        random.seed(42)
        result = ss.NextSplittableSelector.selectRandom(field.cells)
        
        self.assertEqual(result, expected)
        self.assert_(field.cells[result].isSplittable())
    
    def testSelectMostRules(self):
        """
        Test that selectRandom() returns the correct cell with most related
        rules.
        
        """
        field = sg.Field(2,2)
        field.digits[0][0].setFeasibleValues(set(range(1,9)))
        field.digits[0][1].setFeasibleValues(set(range(4)))
        field.digits[1][0].setFeasibleValues(set())
        field.digits[1][1].setFeasibleValues(set(range(10)))
        
        rule = sr.UnaryRule(field.digits[1][1])
        
        random.seed(42)
        result = ss.NextSplittableSelector.selectWithMostRules(field.cells)
        
        self.assertEqual(field.cells[result], field.digits[1][1])
    

# Execute unit tests when called standalone.
if __name__ == "__main__":
    unittest.main()

# vim:sw=4:ts=4:sts=4:expandtab
